from flask import Flask, render_template, request, redirect, url_for, jsonify, flash, session
from database import init_db, get_db
from scraper import run_scraper
import threading
import time
from datetime import datetime
import json
from auth import auth
import psycopg2
from psycopg2.extras import DictCursor

app = Flask(__name__)
app.secret_key = 'powercxkberkkfsdf591s6fdg4'  # You can keep this for flash messages
app.register_blueprint(auth)

@app.template_filter('format_datetime')
def format_datetime(value, format='%Y-%m-%d %H:%M:%S'):
    if value is None:
        return ''
    return datetime.fromtimestamp(value).strftime(format)

@app.template_filter('from_json')
def from_json(value):
    return json.loads(value) if value else []

@app.route('/')
def dashboard():
    if 'user_id' not in session:
        return redirect(url_for('auth.login'))
    db = get_db()
    with db.cursor(cursor_factory=psycopg2.extras.DictCursor) as cur:
        cur.execute('SELECT * FROM scrapers WHERE user_id = %s ORDER BY id ASC', (session['user_id'],))
        scrapers = cur.fetchall()
    return render_template('dashboard.html', scrapers=scrapers)

@app.route('/proxies', methods=['GET', 'POST'])
def manage_proxies():
    if 'user_id' not in session:
        return redirect(url_for('auth.login'))
    
    db = get_db()
    if request.method == 'POST':
        proxies = request.form['proxies'].split('\n')
        with db.cursor() as cur:
            cur.execute('DELETE FROM proxies WHERE user_id = %s', (session['user_id'],))
            for proxy in proxies:
                proxy = proxy.strip()
                if proxy:
                    cur.execute('INSERT INTO proxies (proxy, user_id) VALUES (%s, %s)', (proxy, session['user_id']))
        db.commit()
        flash('Proxies updated successfully')
        return redirect(url_for('manage_proxies'))

    with db.cursor(cursor_factory=psycopg2.extras.DictCursor) as cur:
        cur.execute('SELECT * FROM proxies WHERE user_id = %s', (session['user_id'],))
        proxies = cur.fetchall()
    return render_template('manage_proxies.html', proxies=proxies)

@app.route('/create', methods=['GET', 'POST'])
def create_scraper():
    if 'user_id' not in session:
        return redirect(url_for('auth.login'))
    if request.method == 'POST':
        name = request.form['name']
        url = request.form['url']
        selector_type = request.form['selector_type']
        use_direct_connection = 'use_direct_connection' in request.form
        selector = request.form['selector']
        interval = int(request.form['interval'])
        use_proxy = 'use_proxy' in request.form
        proxy = request.form['proxy'] if use_proxy else None
        use_rotating_proxies = 'use_rotating_proxies' in request.form
        
        db = get_db()
        with db.cursor() as cur:
            cur.execute('INSERT INTO scrapers (name, url, selector_type, selector, interval, use_proxy, proxy, use_rotating_proxies, use_direct_connection, user_id) VALUES (%s, %s, %s, %s, %s, %s, %s, %s, %s, %s)',
                       (name, url, selector_type, selector, interval, use_proxy, proxy, use_rotating_proxies, use_direct_connection, session['user_id']))
        db.commit()
        return redirect(url_for('dashboard'))
    return render_template('create_scraper.html')

@app.route('/edit/<int:id>', methods=['GET', 'POST'])
def edit_scraper(id):
    if 'user_id' not in session:
        return redirect(url_for('auth.login'))
    db = get_db()
    with db.cursor(cursor_factory=psycopg2.extras.DictCursor) as cur:
        cur.execute('SELECT * FROM scrapers WHERE id=%s AND user_id=%s', (id, session['user_id']))
        scraper = cur.fetchone()
    if scraper is None:
        return redirect(url_for('dashboard'))
    if request.method == 'POST':
        name = request.form['name']
        url = request.form['url']
        selector_type = request.form['selector_type']
        use_direct_connection = 'use_direct_connection' in request.form
        selector = request.form['selector']
        interval = int(request.form['interval'])
        use_proxy = 'use_proxy' in request.form
        proxy = request.form['proxy'] if use_proxy else None
        use_rotating_proxies = 'use_rotating_proxies' in request.form
        
        with db.cursor() as cur:
            cur.execute('UPDATE scrapers SET name=%s, url=%s, selector_type=%s, selector=%s, interval=%s, use_proxy=%s, proxy=%s, use_rotating_proxies=%s, use_direct_connection=%s WHERE id=%s',
                       (name, url, selector_type, selector, interval, use_proxy, proxy, use_rotating_proxies, use_direct_connection, id))
        db.commit()
        
        return redirect(url_for('dashboard'))
    
    return render_template('edit_scraper.html', scraper=scraper)

@app.route('/delete/<int:id>')
def delete_scraper(id):
    if 'user_id' not in session:
        return redirect(url_for('auth.login'))
    db = get_db()
    with db.cursor() as cur:
        # First, delete associated results
        cur.execute('DELETE FROM scraper_results WHERE scraper_id = %s', (id,))
        # Then, delete the scraper
        cur.execute('DELETE FROM scrapers WHERE id = %s AND user_id = %s', (id, session['user_id']))
    db.commit()
    return redirect(url_for('dashboard'))

@app.route('/run/<int:id>')
def run_scraper_route(id):
    if 'user_id' not in session:
        return redirect(url_for('auth.login'))
    db = get_db()
    with db.cursor(cursor_factory=psycopg2.extras.DictCursor) as cur:
        cur.execute('SELECT * FROM scrapers WHERE id=%s AND user_id=%s', (id, session['user_id']))
        scraper = cur.fetchone()
    if scraper is None:
        return jsonify({'error': 'Scraper not found'}), 404
    result = run_scraper(scraper['url'], scraper['selector_type'], scraper['selector'], 
                         scraper['use_proxy'], scraper['proxy'], scraper['use_rotating_proxies'])
    
    # Save the results immediately
    with db.cursor() as cur:
        cur.execute('UPDATE scrapers SET last_run=%s WHERE id=%s',
                   (time.time(), id))
        cur.execute('INSERT INTO scraper_results (scraper_id, result, scraped_at) VALUES (%s, %s, %s)',
                   (id, json.dumps(result), time.time()))
    db.commit()
    
    return jsonify(result)

@app.route('/schedules')
def view_schedules():
    if 'user_id' not in session:
        return redirect(url_for('auth.login'))
    db = get_db()
    with db.cursor(cursor_factory=psycopg2.extras.DictCursor) as cur:
        cur.execute('SELECT id, name, interval, last_run FROM scrapers WHERE user_id=%s', (session['user_id'],))
        scrapers = cur.fetchall()
    current_time = time.time()
    
    schedules = []
    for scraper in scrapers:
        next_run = scraper['last_run'] + (scraper['interval'] * 60)
        schedules.append({
            'id': scraper['id'],
            'name': scraper['name'],
            'interval': scraper['interval'],
            'last_run': scraper['last_run'],
            'next_run': next_run,
            'status': 'Running' if next_run <= current_time else 'Scheduled'
        })
    
    return render_template('schedules.html', schedules=schedules)

@app.route('/view/<int:id>')
def view_scraper(id):
    db = get_db()
    with db.cursor(cursor_factory=psycopg2.extras.DictCursor) as cur:
        cur.execute('SELECT * FROM scrapers WHERE id=%s', (id,))
        scraper = cur.fetchone()
    if scraper is None:
        return redirect(url_for('dashboard'))
    
    page = request.args.get('page', 1, type=int)
    per_page = 100
    offset = (page - 1) * per_page
    
    with db.cursor(cursor_factory=psycopg2.extras.DictCursor) as cur:
        cur.execute('SELECT * FROM scraper_results WHERE scraper_id=%s ORDER BY scraped_at DESC LIMIT %s OFFSET %s',
                    (id, per_page, offset))
        results = cur.fetchall()
    
    processed_results = []
    for result in results:
        processed_result = dict(result)
        processed_result['result'] = json.loads(result['result']) if result['result'] else []
        processed_results.append(processed_result)

    with db.cursor() as cur:
        cur.execute('SELECT COUNT(*) FROM scraper_results WHERE scraper_id=%s', (id,))
        total_results = cur.fetchone()[0]
    total_pages = min(50, -(-total_results // per_page))  # Ceiling division, max 50 pages
    
    is_logged_in = 'user_id' in session
    return render_template('view_scraper.html', scraper=scraper, results=processed_results, page=page, total_pages=total_pages, is_logged_in=is_logged_in)

@app.route('/mass_create', methods=['GET', 'POST'])
def mass_create_scrapers():
    if 'user_id' not in session:
        return redirect(url_for('auth.login'))
    
    db = get_db()
    if request.method == 'POST':
        scrapers_data = request.form['scrapers'].split('\n')
        processed_scrapers = []
        
        with db.cursor() as cur:
            for scraper in scrapers_data:
                parts = scraper.strip().split(',')
                if len(parts) == 8:
                    name, url, selector_type, selectors, interval, proxy, rotate_proxy, use_direct_proxy_fail = [p.strip() for p in parts]
                    selectors = selectors.split('|')  # Split multiple selectors
                    proxy = proxy.split('[')[1].split(']')[0] if proxy.startswith('proxy[') else None
                    rotate_proxy = rotate_proxy.split('[')[1].split(']')[0] == 'yes'
                    use_direct_proxy_fail = use_direct_proxy_fail.split('[')[1].split(']')[0] == 'yes'
                    
                    # Check if scraper already exists
                    cur.execute('SELECT id FROM scrapers WHERE name = %s AND user_id = %s', (name, session['user_id']))
                    existing_scraper = cur.fetchone()
                    
                    if existing_scraper:
                        # Update existing scraper
                        cur.execute('''
                            UPDATE scrapers 
                            SET url = %s, selector_type = %s, selector = %s, interval = %s, 
                                use_proxy = %s, proxy = %s, use_rotating_proxies = %s, use_direct_connection = %s
                            WHERE id = %s
                        ''', (url, selector_type, '|'.join(selectors), int(interval), bool(proxy), proxy, rotate_proxy, use_direct_proxy_fail, existing_scraper[0]))
                    else:
                        # Insert new scraper
                        cur.execute('''
                            INSERT INTO scrapers 
                            (name, url, selector_type, selector, interval, use_proxy, proxy, use_rotating_proxies, use_direct_connection, user_id) 
                            VALUES (%s, %s, %s, %s, %s, %s, %s, %s, %s, %s)
                        ''', (name, url, selector_type, '|'.join(selectors), int(interval), bool(proxy), proxy, rotate_proxy, use_direct_proxy_fail, session['user_id']))
                    
                    processed_scrapers.append(scraper)
        
        db.commit()
        flash('Scrapers updated successfully')
        return render_template('mass_create.html', scrapers='\n'.join(processed_scrapers))
    
    # GET request: fetch existing scrapers
    with db.cursor(cursor_factory=psycopg2.extras.DictCursor) as cur:
        cur.execute('SELECT * FROM scrapers WHERE user_id = %s', (session['user_id'],))
        scrapers = cur.fetchall()
    
    scrapers_text = '\n'.join([
        f"{s['name']},{s['url']},{s['selector_type']},{s['selector'].replace('|', '|')},{s['interval']},proxy[{s['proxy'] or ''}],rotate_proxy[{'yes' if s['use_rotating_proxies'] else 'no'}],use_direct_proxy_fail[{'yes' if s['use_direct_connection'] else 'no'}]"
        for s in scrapers
    ])
    
    return render_template('mass_create.html', scrapers=scrapers_text)

@app.route('/mass_view')
def mass_view():
    db = get_db()
    with db.cursor(cursor_factory=psycopg2.extras.DictCursor) as cur:
        cur.execute('''
            SELECT s.id, s.name, sr.result, sr.scraped_at
            FROM scrapers s
            LEFT JOIN (
                SELECT scraper_id, result, scraped_at,
                    ROW_NUMBER() OVER (PARTITION BY scraper_id ORDER BY scraped_at DESC) as rn
                FROM scraper_results
            ) sr ON s.id = sr.scraper_id AND sr.rn = 1
            ORDER BY s.id ASC
        ''')
        scrapers = cur.fetchall()
    
    return render_template('mass_view.html', scrapers=scrapers)

@app.route('/create_group', methods=['GET', 'POST'])
def create_group():
    if 'user_id' not in session:
        return redirect(url_for('auth.login'))
    
    db = get_db()
    if request.method == 'POST':
        group_name = request.form['group_name']
        scrapers_data = request.form['scrapers'].split('\n')
        
        with db.cursor() as cur:
            cur.execute('INSERT INTO scraper_groups (name, user_id) VALUES (%s, %s) RETURNING id',
                        (group_name, session['user_id']))
            group_id = cur.fetchone()[0]
            
            for scraper in scrapers_data:
                parts = scraper.strip().split(',')
                if len(parts) == 8:
                    name, url, selector_type, selectors, interval, proxy, rotate_proxy, use_direct_proxy_fail = [p.strip() for p in parts]
                    selectors = selectors.split('|')
                    proxy = proxy.split('[')[1].split(']')[0] if proxy.startswith('proxy[') else None
                    rotate_proxy = rotate_proxy.split('[')[1].split(']')[0] == 'yes'
                    use_direct_proxy_fail = use_direct_proxy_fail.split('[')[1].split(']')[0] == 'yes'
                    
                    cur.execute('''
                        INSERT INTO scrapers 
                        (name, url, selector_type, selector, interval, use_proxy, proxy, use_rotating_proxies, use_direct_connection, user_id, group_id) 
                        VALUES (%s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s)
                    ''', (name, url, selector_type, '|'.join(selectors), int(interval), bool(proxy), proxy, rotate_proxy, use_direct_proxy_fail, session['user_id'], group_id))
        
        db.commit()
        flash('Scraper group created successfully')
        return redirect(url_for('view_group', group_id=group_id))
    
    return render_template('create_group.html')

@app.route('/view_group/<int:group_id>')
def view_group(group_id):
    db = get_db()
    with db.cursor(cursor_factory=psycopg2.extras.DictCursor) as cur:
        cur.execute('SELECT * FROM scraper_groups WHERE id = %s', (group_id,))
        group = cur.fetchone()
        
        if group is None:
            flash('Group not found')
            return redirect(url_for('dashboard'))
        
        cur.execute('''
            SELECT s.*, sr.result, sr.scraped_at
            FROM scrapers s
            LEFT JOIN (
                SELECT scraper_id, result, scraped_at,
                    ROW_NUMBER() OVER (PARTITION BY scraper_id ORDER BY scraped_at DESC) as rn
                FROM scraper_results
            ) sr ON s.id = sr.scraper_id AND sr.rn = 1
            WHERE s.group_id = %s
            ORDER BY s.id ASC
        ''', (group_id,))
        scrapers = cur.fetchall()
    
    return render_template('view_group.html', group=group, scrapers=scrapers)


@app.route('/edit_group/<int:group_id>', methods=['GET', 'POST'])
def edit_group(group_id):
    if 'user_id' not in session:
        return redirect(url_for('auth.login'))
    
    db = get_db()
    with db.cursor(cursor_factory=psycopg2.extras.DictCursor) as cur:
        cur.execute('SELECT * FROM scraper_groups WHERE id = %s AND user_id = %s', (group_id, session['user_id']))
        group = cur.fetchone()
        
        if group is None:
            flash('Group not found')
            return redirect(url_for('dashboard'))
        
        if request.method == 'POST':
            group_name = request.form['group_name']
            scrapers_data = request.form['scrapers'].split('\n')
            
            cur.execute('UPDATE scraper_groups SET name = %s WHERE id = %s', (group_name, group_id))
            cur.execute('DELETE FROM scrapers WHERE group_id = %s', (group_id,))
            
            for scraper in scrapers_data:
                parts = scraper.strip().split(',')
                if len(parts) == 8:
                    name, url, selector_type, selectors, interval, proxy, rotate_proxy, use_direct_proxy_fail = [p.strip() for p in parts]
                    selectors = selectors.split('|')
                    proxy = proxy.split('[')[1].split(']')[0] if proxy.startswith('proxy[') else None
                    rotate_proxy = rotate_proxy.split('[')[1].split(']')[0] == 'yes'
                    use_direct_proxy_fail = use_direct_proxy_fail.split('[')[1].split(']')[0] == 'yes'
                    
                    cur.execute('''
                        INSERT INTO scrapers 
                        (name, url, selector_type, selector, interval, use_proxy, proxy, use_rotating_proxies, use_direct_connection, user_id, group_id) 
                        VALUES (%s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s)
                    ''', (name, url, selector_type, '|'.join(selectors), int(interval), bool(proxy), proxy, rotate_proxy, use_direct_proxy_fail, session['user_id'], group_id))
            
            db.commit()
            flash('Group updated successfully')
            return redirect(url_for('view_group', group_id=group_id))
        
        cur.execute('SELECT * FROM scrapers WHERE group_id = %s', (group_id,))
        scrapers = cur.fetchall()
    
    scrapers_text = '\n'.join([
        f"{s['name']},{s['url']},{s['selector_type']},{s['selector'].replace('|', '|')},{s['interval']},proxy[{s['proxy'] or ''}],rotate_proxy[{'yes' if s['use_rotating_proxies'] else 'no'}],use_direct_proxy_fail[{'yes' if s['use_direct_connection'] else 'no'}]"
        for s in scrapers
    ])
    
    return render_template('edit_group.html', group=group, scrapers=scrapers_text)

@app.route('/view_group_list')
def view_group_list():
    if 'user_id' not in session:
        return redirect(url_for('auth.login'))
    
    db = get_db()
    with db.cursor(cursor_factory=psycopg2.extras.DictCursor) as cur:
        cur.execute('''
            SELECT sg.id, sg.name, COUNT(s.id) as scraper_count
            FROM scraper_groups sg
            LEFT JOIN scrapers s ON sg.id = s.group_id
            WHERE sg.user_id = %s
            GROUP BY sg.id, sg.name
            ORDER BY sg.id ASC
        ''', (session['user_id'],))
        groups = cur.fetchall()
    
    return render_template('view_group_list.html', groups=groups)

@app.route('/delete_group/<int:group_id>')
def delete_group(group_id):
    if 'user_id' not in session:
        return redirect(url_for('auth.login'))
    
    db = get_db()
    with db.cursor() as cur:
        cur.execute('DELETE FROM scrapers WHERE group_id = %s', (group_id,))
        cur.execute('DELETE FROM scraper_groups WHERE id = %s AND user_id = %s', (group_id, session['user_id']))
    db.commit()
    flash('Group deleted successfully')
    return redirect(url_for('view_group_list'))

@app.route('/run_group/<int:group_id>')
def run_group(group_id):
    if 'user_id' not in session:
        return redirect(url_for('auth.login'))
    
    db = get_db()
    with db.cursor(cursor_factory=psycopg2.extras.DictCursor) as cur:
        cur.execute('SELECT * FROM scrapers WHERE group_id = %s', (group_id,))
        scrapers = cur.fetchall()
    
    results = []
    for scraper in scrapers:
        result = run_scraper(
            scraper['url'],
            scraper['selector_type'],
            scraper['selector'],
            scraper['use_proxy'],
            scraper['proxy'],
            scraper['use_rotating_proxies'],
            scraper['use_direct_connection']
        )
        results.append({'scraper_id': scraper['id'], 'result': result})
        
        with db.cursor() as cur:
            cur.execute('UPDATE scrapers SET last_run = %s WHERE id = %s', (time.time(), scraper['id']))
            cur.execute('INSERT INTO scraper_results (scraper_id, result, scraped_at) VALUES (%s, %s, %s)',
                       (scraper['id'], json.dumps(result), time.time()))
    
    db.commit()
    flash('Group scrapers run successfully')
    return redirect(url_for('view_group', group_id=group_id))


def scraper_thread():
    while True:
        try:
            with app.app_context():
                db = get_db()
                with db.cursor(cursor_factory=psycopg2.extras.DictCursor) as cur:
                    cur.execute('SELECT * FROM scrapers')
                    scrapers = cur.fetchall()
                
                current_time = time.time()
                next_run = float('inf')
                
                for scraper in scrapers:
                    last_run = scraper['last_run'] or 0  # Use 0 if last_run is None
                    time_since_last_run = current_time - last_run
                    if time_since_last_run >= scraper['interval'] * 60:
                        print(f"Running scraper {scraper['id']}")
                        result = run_scraper(
                            scraper['url'],
                            scraper['selector_type'],
                            scraper['selector'],
                            scraper['use_proxy'],
                            scraper['proxy'],
                            scraper['use_rotating_proxies'],
                            scraper['use_direct_connection']
                        )
                        print(f"Scraper {scraper['id']} result: {result}")
                        with db.cursor() as cur:
                            cur.execute('UPDATE scrapers SET last_run=%s WHERE id=%s',
                                       (current_time, scraper['id']))
                            cur.execute('INSERT INTO scraper_results (scraper_id, result, scraped_at) VALUES (%s, %s, %s)',
                                       (scraper['id'], json.dumps(result), current_time))
                        db.commit()
                        next_run = min(next_run, scraper['interval'] * 60)
                        time.sleep(1)  # Add a small delay after each scraper run
                    else:
                        next_run = min(next_run, scraper['interval'] * 60 - time_since_last_run)
                
                sleep_time = max(1, min(next_run, 60))  # Sleep at least 1 second, at most 60 seconds
                time.sleep(sleep_time)
        except Exception as e:
            print(f"An error occurred: {e}")
            time.sleep(60)

if __name__ == '__main__':
    init_db()
    threading.Thread(target=scraper_thread, daemon=True).start()
    app.run(host="0.0.0.0", port=5001, debug=True)

