from flask import Blueprint, render_template, request, redirect, url_for, flash, session, abort
from werkzeug.security import generate_password_hash, check_password_hash
from database import get_db
from functools import wraps
import time
import psycopg2
from psycopg2.extras import DictCursor

LOGIN_ATTEMPTS = {}
BLOCKED_IPS = {}
MAX_ATTEMPTS = 5
BLOCK_TIME = 600  # 5 minutes in seconds

def check_blocked_ip(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        ip = request.remote_addr
        if ip in BLOCKED_IPS:
            if time.time() - BLOCKED_IPS[ip] < BLOCK_TIME:
                abort(403)  # Forbidden
            else:
                del BLOCKED_IPS[ip]
        return f(*args, **kwargs)
    return decorated_function


auth = Blueprint('auth', __name__)

@auth.route('/register', methods=['GET', 'PcOST'])
def register():
    if request.method == 'POST':
        username = request.form['username']
        password = request.form['password']
        db = get_db()
        error = None

        if not username:
            error = 'Username is required.'
        elif not password:
            error = 'Password is required.'
        else:
            with db.cursor() as cur:
                cur.execute('SELECT id FROM users WHERE username = %s', (username,))
                if cur.fetchone() is not None:
                    error = 'User {} is already registered.'.format(username)

        if error is None:
            with db.cursor() as cur:
                cur.execute('INSERT INTO users (username, password) VALUES (%s, %s)',
                           (username, generate_password_hash(password)))
            db.commit()
            return redirect(url_for('auth.login'))

        flash(error)

    return render_template('register.html')

@auth.route('/login', methods=['GET', 'POST'])
@check_blocked_ip
def login():
    if request.method == 'POST':
        username = request.form['username']
        password = request.form['password']
        db = get_db()
        error = None
        
        with db.cursor(cursor_factory=psycopg2.extras.DictCursor) as cur:
            cur.execute('SELECT * FROM users WHERE username = %s', (username,))
            user = cur.fetchone()

        ip = request.remote_addr
        if ip not in LOGIN_ATTEMPTS:
            LOGIN_ATTEMPTS[ip] = {'count': 0, 'time': time.time()}

        if user is None:
            error = 'Incorrect username.'
        elif not check_password_hash(user['password'], password):
            error = 'Incorrect password.'

        if error is None:
            session.clear()
            session['user_id'] = user['id']
            session['username'] = user['username']
            LOGIN_ATTEMPTS[ip]['count'] = 0
            return redirect(url_for('dashboard'))

        LOGIN_ATTEMPTS[ip]['count'] += 1
        if LOGIN_ATTEMPTS[ip]['count'] >= MAX_ATTEMPTS:
            BLOCKED_IPS[ip] = time.time()
            del LOGIN_ATTEMPTS[ip]
            flash('Too many failed attempts. Your IP has been blocked for 5 minutes.')
            return redirect(url_for('auth.login'))

        flash(error)

    return render_template('login.html')

@auth.route('/logout')
def logout():
    session.clear()
    return redirect(url_for('auth.login'))
